#include <stdlib.h>
#include <ctype.h>
#include <tcl.h>
#include <stdio.h>
#include <math.h>
#include <string.h>		/* strcat */

#include "rpm-internal.h"

static char rcsid[] = "$Id: rpmCmd.c,v 1.18 1999/08/09 13:51:11 martin Exp $";

static int cinit = 0;
/*
**	tcl handle support for rpm's
*/
static Tcl_HashTable rpmHT;
static int rpmHTReady = 0;

static void initHT()
{
    if(rpmHTReady)
		return;
    Tcl_InitHashTable(&rpmHT, TCL_STRING_KEYS);
    ++rpmHTReady;
}
/*
**	return a new handle name in static memory
*/
static char* rpmNewHandle()
{
    static long nh = 0;
    static char name[64];

    sprintf(name, ":rpm%ld", nh++);
    return name;
}
/*
**	load a vB into the hash table.
**	If the name already exists overwrite its value.
**	If name is NULL, create an auto name.
**	In either event return a pointer to name (which may
**		be in static memory).
**
**      if the new value is dynamic, we use its storage,
**          otherwise use a deep copy.
*/
char* rpmAsHandle(char* name, rpmBinary vb)
{
    char* rn;
    Tcl_HashEntry* he;
    int newhe;
    rpmBinary* vp;

    initHT();
    rn = name;
    if(rn == 0)
		rn = rpmNewHandle();
    he = Tcl_CreateHashEntry(&rpmHT, rn, &newhe);
    vp = (rpmBinary *) he->clientData;
    if(newhe == 0) {
		if(vp->dynamic != 0)
			free(vp->v);;
    } else {
		vp = (rpmBinary*) malloc(sizeof(rpmBinary));
    }
    he->clientData = vp;
    *vp = vb;
    if(vb.dynamic == 0)
		*vp = rpmDeepCopy(vb);
    vb.dynamic = 0;
    vp->dynamic = 1;
    vp->good = 1;
    return rn;
}
/*
**	return the vB with an existing name or 0 if it doesn't
**	exist.
*/
rpmBinary* rpmGetHandle(char* name)
{
    Tcl_HashEntry* he;

    initHT();
    he = Tcl_FindHashEntry(&rpmHT, name);
    if(he == 0)
		return 0;
    return (rpmBinary *) he->clientData;
}
/*
**	trash a handle if it exists
*/
int rpmDeleteHandle(char* name)
{
    Tcl_HashEntry* he;
    rpmBinary* vp;

    initHT();
    he = Tcl_FindHashEntry(&rpmHT, name);
    if(he == 0)
		return TCL_ERROR;
    vp = (rpmBinary *) he->clientData;
    rpmFreeBinary(*vp);
    free(vp);
    Tcl_DeleteHashEntry(he);
    return TCL_OK;
}

static rpmArgType argIs(char* arg)
{
    while(isspace(*arg))
		++arg;

    if(*arg == '\'')
		return rString;
    if(*arg == '@')
		return rHandle;
    if(*arg == '{')
		return rNumbers;
    if((*arg == '-') || (*arg == '+'))
		arg++;
    if((*arg) && (isdigit(*arg)))
		return rNumbers;
    if((*arg) && (*arg == '.'))
		return rNumbers;
    return rCommand;
    
}

static int rpmPushArgAsBinary(Tcl_Interp* interp, char* arg)
{
    rpmBinary v;
    rpmBinary* vptr;

    switch(argIs(arg)) {
    case rHandle:
		arg++;
		vptr = rpmGetHandle(arg);
		if(vptr == 0) {
			Tcl_AppendResult(interp, "bad handle: ", arg, 0);
			return TCL_ERROR;
		}
		v = *vptr;
		v.dynamic = 0;
		break;
    case rString:
		arg++;
		v = rpmNewBinary(0,0);
		v.usedas =rString;
		v.string = malloc(strlen(arg) + 1);
		strcpy(v.string, arg);
		break;
    case rNumbers:
		v = rpmAtof(interp, arg);
		v.good = 1;
		break;
    default:
		Tcl_AppendResult(interp, "bad arg type for: ", arg, 0);
		return TCL_ERROR;
    }
    rpmPushB(v);
    return TCL_OK;
}

static void rpmBinary2Tcl(Tcl_Interp* interp, rpmBinary v,
						  int level, int offs)
{
    int i;
    char q[128];
    int l, j;

    if(cinit)
		fprintf(stderr, "bin2tcl: level %d  offs %d\n", level, offs);

    if(v.usedas == rString) {
		Tcl_AppendResult(interp, v.string, 0);
		return;
    }
    if(v.usedas == rHandle) {
		Tcl_AppendResult(interp, v.hname, 0);
		return;
    }
    if(level == (v.dimens - 1)) {
		for(i = 0; i < v.dim[level]; i++) {
			sprintf(q, "%g", v.v[offs + i]);
			Tcl_AppendResult(interp, q, 0);
			if(i < (v.dim[level] - 1))
				Tcl_AppendResult(interp, " ", 0);
		}
		return;
    }
    l = 1;
    for(j = level + 1; j < v.dimens; j++)
		l *= v.dim[j];
    for(i = 0; i < v.dim[level]; i++) {
		Tcl_AppendResult(interp, "{ ", 0);
		rpmBinary2Tcl(interp, v, level + 1, offs + l * i);
		Tcl_AppendResult(interp, " } ", 0);
    }
}

/*
**	bad args error message
*/
int rpmBadArgs(Tcl_Interp* interp, int argc, char** argv)
{
    Tcl_AppendResult(interp, "bad rpm command: should be rpm ",
					 (char *) NULL);
    return TCL_ERROR;
}
/*
**	general purpose error message
*/
int rpmBadNews(Tcl_Interp* interp, char* s)
{
    Tcl_AppendResult(interp, s, (char *) NULL);
    return TCL_ERROR;
}
/*
**	pop
*/
static int rpmPopSubCmd(ClientData dummy, Tcl_Interp* interp)
{
    if(rpmStackDepth() > 0)
		rpmFreeBinary(rpmPop());
    return TCL_OK;
}
/*
**	dup
*/
static int rpmDupSubCmd(ClientData dummy, Tcl_Interp* interp)
{
    rpmBinary a, b;
    if(rpmStackDepth() > 0) {
		a = rpmPop();
		b = rpmDeepCopy(a);
		rpmPushB(a);
		rpmPushB(b);
    }
    return TCL_OK;
}
/*
**	swap
*/
static int rpmSwapSubCmd(ClientData dummy, Tcl_Interp* interp)
{
    rpmBinary a, b;
    if(rpmStackDepth() > 1) {
		b = rpmPop();
		a = rpmPop();
		rpmPushB(b);
		rpmPushB(a);
    }
    return TCL_OK;
}

/*
**	[name] handle
*/
static int rpmHandleSubCmd(ClientData dummy, Tcl_Interp* interp)
{
	rpmBinary a;
	rpmBinary b;
	char* name;
	char* s;

	if(rpmStackDepth() < 1) {
		Tcl_AppendResult(interp, "rpm handle: empty stack.", 0);
		return TCL_ERROR;
	}

	a = rpmPop();
	if(a.usedas == rString) {
		if(rpmStackDepth() < 1) {
			Tcl_AppendResult(interp,
							 "rpm handle: empty stack after string.", 0);
			return TCL_ERROR;
		}
		b = a;
		name = b.string;
		a = rpmPop();
	} else
		name = 0;
	s = rpmAsHandle(name, a);
	if(name != 0)
		rpmFreeBinary(b);
	b.usedas = rString;
	b.dynamic = 1;
	b.string = malloc(strlen(s) + 1);
	strcpy(b.string, s);
	rpmPushB(b);
	return TCL_OK;
}

/*
**	return info about the stack\'s contents
*/
static int rpmInfoSubCmd(ClientData dummy, Tcl_Interp* interp)
{
    int st;
    rpmBinary v;
    int i;
    char q[128];
    double l;
    double ymax, ymin, ybar;

    st = 0;
    while(rpmStackDepth() > 0) {
		if(st > 0)
			Tcl_AppendResult(interp, "\n", 0);
		v = rpmPop();
		switch(v.usedas) {
		case rHandle:
			Tcl_AppendResult(interp, v.hname, " (handle)", 0);
			break;
		case rString:
			Tcl_AppendResult(interp, v.string, " (string)", 0);
			break;
		case rNumbers:
			sprintf(q, "%d", rpmLength(v));
			Tcl_AppendResult(interp, q, " element(s) as ", 0);
			for(i = 0; i < v.dimens; i++) {
				sprintf(q, "%d", v.dim[i]);
				Tcl_AppendResult(interp, q,
								 (i == (v.dimens - 1)) ? "" : "x", 0);
			}
			ymin = ymax = ybar = v.v[0];
			l = rpmLength(v);
			for(i = 1; i < l; i++) {
				if(v.v[i] > ymax)
					ymax = v.v[i];
				if(v.v[i] < ymin)
					ymin = v.v[i];
				ybar += v.v[i];
			}
			ybar /= l;
			sprintf(q, "  %g >= y >= %g  ybar = %g", ymax, ymin, ybar);
			Tcl_AppendResult(interp, q, 0);
			break;
		default:
			break;
		}
		rpmFreeBinary(v);
		st++;
    }
    return TCL_OK;
}

/*
**	version
*/
static int rpmVersionSubCmd(ClientData dummy, Tcl_Interp* interp)
{
    rpmBinary s;

    s.usedas = rString;
    s.string = malloc(strlen(RpmVersion) + 1);
    strcpy(s.string, RpmVersion);
    rpmPushB(s);
    return TCL_OK;
}
/*
**	names
*/
static int rpmNamesSubCmd(ClientData dummy, Tcl_Interp* interp)
{
    Tcl_HashEntry* he;
    Tcl_HashSearch srch;
    char* name;
    int n;

    he = Tcl_FirstHashEntry(&rpmHT, &srch);
    for(n = 0; 1; n++) {
		if(he == 0)
			return TCL_OK;
		name = Tcl_GetHashKey(&rpmHT, he);
		if(n > 0)
			Tcl_AppendResult(interp, " ", (char *) NULL);
		Tcl_AppendResult(interp, name, (char *) NULL);
		he = Tcl_NextHashEntry(&srch);
    }
    return TCL_OK;
}
/*
**	name delete
*/
static int rpmDeleteSubCmd(ClientData dummy, Tcl_Interp* interp)
{
    rpmBinary h;

    h = rpmPop();
    if(h.usedas != rString) {
		Tcl_AppendResult(interp, "rpm delete: arg not a string", 0);
		return TCL_ERROR;
    }
    if(rpmDeleteHandle(h.string) == TCL_ERROR) {
		Tcl_AppendResult(interp, "rpm delete: can't zap ", h.string, 0);
		rpmFreeBinary(h);
		return TCL_ERROR;
    }
    rpmFreeBinary(h);
    return TCL_OK;
}
/*
**	oldname newname rename
*/
static int rpmRenameSubCmd(ClientData dummy, Tcl_Interp* interp)
{
    rpmBinary new;
    rpmBinary old;

    Tcl_HashEntry* he;
    rpmBinary* rp;
    int isnew;

    new = rpmPop();
    old = rpmPop();
    if(new.usedas != rString) {
		Tcl_AppendResult(interp, "rpm rename: new not a string", 0);
		return TCL_ERROR;
    }
    if(old.usedas != rString) {
		Tcl_AppendResult(interp, "rpm rename: old not a string", 0);
		return TCL_ERROR;
    }

    he = Tcl_FindHashEntry(&rpmHT, old.string);
    if(he == 0) {
		Tcl_AppendResult(interp, "rpm rename: old not a handle", 0);
		return TCL_ERROR;
    }
    rp = he->clientData;
    Tcl_DeleteHashEntry(he);
    he = Tcl_CreateHashEntry(&rpmHT, new.string, &isnew);
    if(isnew != 0) {
		rpmFreeBinary(* (rpmBinary *) he->clientData);
		free(he->clientData);
    }
    he->clientData = rp;
    rpmFreeBinary(new);
    rpmFreeBinary(old);
    return TCL_OK;
}

typedef int ((*subProc)(ClientData, Tcl_Interp*));

typedef struct {
    char*	subname;
    subProc	f;
    int		minStack;
} subEntry;



static subEntry subTable[] = {

    {"pop", rpmPopSubCmd, 1},
    {"dup", rpmDupSubCmd, 1},
    {"swap", rpmSwapSubCmd, 2},
    {"cat", rpmCatSubCmd, 2},
    {"flatten", rpmFlattenSubCmd, 1},
    {"transpose", rpmTransposeSubCmd, 1},
    {"flip", rpmFlipSubCmd, 1},
    {"reshape", rpmReshapeSubCmd, 2},
    {"expand", rpmExpandSubCmd, 2},
    {"sexpand", rpmSExpandSubCmd, 2},
    {"shrink", rpmShrinkSubCmd, 2},
    {"handle", rpmHandleSubCmd, 1},

    {"eq", rpmEqSubCmd, 2},
    {"greater", rpmGreaterSubCmd, 2},
    {"lesser", rpmLesserSubCmd, 2},

    {"info", rpmInfoSubCmd, 0},
    {"version", rpmVersionSubCmd, 0},
    {"names",	rpmNamesSubCmd, 0},
    {"delete",	rpmDeleteSubCmd, 1},
    {"rename",	rpmRenameSubCmd, 2},

    {"+", rpmAddSubCmd, 2},
    {"-", rpmSubtractSubCmd, 2},
    {"*", rpmMultiplySubCmd, 2},
    {"/", rpmDivideSubCmd, 2},
    {"sqr", rpmSqrSubCmd, 1},

    {"==", rpmDeqSubCmd, 2},
    {"!=", rpmNeqSubCmd, 2},
    {">", rpmGtSubCmd, 2},
    {">=", rpmGeSubCmd, 2},
    {"<", rpmLtSubCmd, 2},
    {"<=", rpmLeSubCmd, 2},
    {"<=>", rpmCmpSubCmd, 2},
    {"!", rpmNotSubCmd, 2},

    {"length", rpmLengthSubCmd, 1},
    {"range", rpmRangeSubCmd, 3},
    {"min", rpmMinSubCmd, 1},
    {"max", rpmMaxSubCmd, 1},
    {"stats", rpmStatsSubCmd, 1},
    {"sort", rpmSortSubCmd, 1},
    {"sum", rpmSumSubCmd, 1},
    {"psum", rpmPsumSubCmd, 2},
    {"sumsq", rpmSumsqSubCmd, 1},
    {"threshold", rpmThresholdSubCmd, 2},

    {"scaleto", rpmScaletoSubCmd, 3},
    {"abscissate", rpmAbscissateSubCmd, 3},
    {"decimate", rpmDecimateSubCmd, 2},
    {"runavg", rpmRunavgSubCmd, 2},
    {"trssh", rpm3rsshSubCmd, 1},
    {"toplot", rpmToplotSubCmd, 9},

    {"sequence", rpmSequenceSubCmd, 3},
    {"urandom", rpmUrandomSubCmd, 1},
    {"replicate", rpmReplicateSubCmd, 2},

    {"index", rpmIndexSubCmd, 2},
    {"select", rpmSelectSubCmd, 2},

    {"floor", rpmFloorSubCmd, 1},
    {"ceil", rpmCeilSubCmd, 1},
    {"abs", rpmAbsSubCmd, 1},
    {"rint", rpmRintSubCmd, 1},
    {"exp", rpmExpSubCmd, 1},
    {"log", rpmLogSubCmd, 1},
    {"log10", rpmLog10SubCmd, 1},
    {"sqrt", rpmSqrtSubCmd, 1},
    {"sin", rpmSinSubCmd, 1},
    {"cos", rpmCosSubCmd, 1},
    {"tan", rpmTanSubCmd, 1},
    {"asin", rpmAsinSubCmd, 1},
    {"acos", rpmAcosSubCmd, 1},
    {"atan", rpmAtanSubCmd, 1},
    {"mod", rpmModSubCmd, 1},
    {"remainder", rpmRemainderSubCmd, 1},
    {"atan2", rpmAtan2SubCmd, 1},

    {"r2c", rpmR2cCmd, 1},
    {"real", rpmRealCmd, 1},
    {"imag", rpmImagCmd, 1},
    {"dft", rpmDftCmd, 1},
    {"idft", rpmIdftCmd, 1},
    {"shift", rpmShiftDftCmd, 2},
    {"tfilter", rpmTfilterCmd, 2},
    {"getreal", rpmGetrealCmd, 1},

    {"dimensions", rpmDimensionsSubCmd, 1},
    {"get", rpmGetSubCmd, 2},
    {"put", rpmPutSubCmd, 3},

    { 0, 0}
};
/*
**	rpm master command
*/
static int rpmCmd(ClientData dummy, Tcl_Interp* interp, int argc, char** argv)
{
    int ca;
    int i;
    int retv;
    rpmBinary a;

	Tcl_ResetResult(interp);

    if (argc < 2)
		return TCL_OK;

    rpmNewStack();
    for(ca = 1; ca < argc; ca++) {
		switch(argIs(argv[ca])) {
		case rHandle:
		case rNumbers:
		case rString:
			retv = rpmPushArgAsBinary(interp, argv[ca]);
			break;
		case rCommand:
			retv = TCL_OK;
			for(i = 0; subTable[i].subname != 0; i++)
				if(strcmp(argv[ca], subTable[i].subname) == 0) {
					if(rpmStackDepth() < subTable[i].minStack)
						return rpmBadNews(interp, "not enough args");
					retv = subTable[i].f(dummy, interp);
					break;
				}
			if(subTable[i].subname == 0) {
				Tcl_AppendResult(interp, "bad rpm command: ", argv[ca], 0);
				return TCL_ERROR;
			}
			break;
		default:
			break;
		}
		if(retv == TCL_ERROR)
			return retv;
	}
    if(rpmStackDepth() < 1) {
		rpmDeleteStack();
		return TCL_OK;
    }
    a = rpmPop();
    rpmDeleteStack();
    if(!a.good)
		return rpmBadNews(interp, "rpm stack: not a good value");
    rpmBinary2Tcl(interp, a, 0, 0);
    rpmFreeBinary(a);
    return TCL_OK;
}

int Rpm_Init(Tcl_Interp* interp)
{
    Tcl_CreateCommand(interp, "rpm", rpmCmd, (ClientData) 0,
					  (void (*)()) NULL);
    return TCL_OK;
}
